
/* S2 cpu micro architecture simulator

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University
   
   modify from s1 sim Computer Architecture 1997, 1998

   5 December 2001

   s21  						18 Jan 2007
   refresh for s2.3	embed sys	5 Jan 2013
   dual core s2.3x				18 Feb 2013
*/

#include "s23.h"

#define   retvalR	27				// returned value register

int R[2][32], Ir[2], Pc[2], M[MAXMEM];
int runflag[2], savePc[2];
//extern char *cp2;					// output buffer
int DP;								// current free in heap

int int_flag, int_mask;
/*
int signx2( int d ){	// sign bit 21 extended
	if( d & 0x0200000 ) return d | 0xFFC00000;
	return d;
}
*/
//  allocate n words from heap 
PRIVATE int alloc(int n){
	int a;
	a = DP;
	if( DP + n >= MAXMEM ) return 0;
	DP += n;
	return a;
}

PRIVATE int timer1 = 1000;
PRIVATE int divn = 1000;

// simulate a timer
void timer(int flag){
	static int tm;
	static int clk;
	if(flag){			// init
		tm = timer1;
		clk = 0;
	}
	clk++;
	if( clk >= divn ){
		clk = 0;
		tm--;
		if( tm <= 0 ){
			int_flag = 1;	// request interrupt
			tm = timer1;	// reload counter			
		}
	}
}

void trap(int k, int arg, int num ){	// special function
	int n, c;
	switch( num ) {
	case 0:  // stop
		printf("stop, core %d execute %d inst.\n",k,ninst[k]);
		runflag[k] = 0;
		break;
	case 1: // print integer
//		n = sprintf(cp2,"%d", arg);
//		cp2 += n;
		printf("%d",arg);
		break;
	case 2: // printc
//		n = sprintf(cp2,"%c", arg);
//		cp2 += n;
		printf("%c",arg);
		break;
	case 3: // print string
		while( (c = M[arg]) != 0){
			printf("%c",c);
//			*cp2 = (char)c;
			arg++;
//			cp2++;
		}
		break;
/*
	case 10: 
		timer1 = arg;
		timer(0); 
		break; 									// set timer
	case 11: divn = arg; break;					// set divn
	case 12: int_mask = arg; break;				// set int mask
	case 13: int_flag = 0; break;				// clear int
*/
	case 14: R[k][retvalR] = alloc(arg); break;	// alloc
	case 15:                                    // start core 2
		Pc[1] = M[20];                          // link address
		runflag[1] = 1;
		break;
	}
}

void run(int k){		// execute one instruction on core k
	int op, a1, a2, a3;
	savePc[k] = Pc[k];
	Ir[k] = M[Pc[k]];
	Pc[k]++;
	op = IRop(k);
	a1 = IRr1(k);
	a2 = IRr2(k);
	a3 = IRd(k);
//	countclock(op);
	ninst[k]++;
	switch( op ) {
	case LD: 	R[k][a1] = M[a3]; break;
	case LDD: 	R[k][a1] = M[ R[k][a2] + a3]; break;
	case ST: 	M[a3] = R[k][a1]; break;
	case STD: 	M[ R[k][a2] + a3] = R[k][a1]; break;

	case JAL: 	R[k][a1] = Pc[k]; Pc[k] = a3; break;
	case JT: 	if( R[k][a1] != 0 ) Pc[k] = a3; break;
	case JF: 	if( R[k][a1] == 0 ) Pc[k] = a3; break;
	case ADDI: 	R[k][a1] = R[k][a2] + a3; break;
	case SUBI: 	R[k][a1] = R[k][a2] - a3; break;
	case MULI: 	R[k][a1] = R[k][a2] * a3; break;
	case DIVI: 	R[k][a1] = R[k][a2] / a3; break;
	case ANDI: 	R[k][a1] = R[k][a2] & a3; break;
	case ORI:  	R[k][a1] = R[k][a2] | a3; break;
	case XORI: 	R[k][a1] = R[k][a2] ^ a3; break;
	case EQI: 	R[k][a1] = R[k][a2] == a3; break;
	case NEI: 	R[k][a1] = R[k][a2] != a3; break;
	case LTI: 	R[k][a1] = R[k][a2] <  a3; break;
	case LEI: 	R[k][a1] = R[k][a2] <= a3; break;
	case GTI: 	R[k][a1] = R[k][a2] >  a3; break;
	case GEI:  	R[k][a1] = R[k][a2] >= a3; break;
	case SHLI: 	R[k][a1] = R[k][a2] << a3; break;
	case SHRI: 	R[k][a1] = R[k][a2] >> a3; break;
	
	case ADD: 	R[k][a1] = R[k][a2] + R[k][a3]; break;
	case SUB: 	R[k][a1] = R[k][a2] - R[k][a3]; break;
	case MUL: 	R[k][a1] = R[k][a2] * R[k][a3]; break;
	case DIV: 	R[k][a1] = R[k][a2] / R[k][a3]; break;
	case AND: 	R[k][a1] = R[k][a2] & R[k][a3]; break;
	case OR:  	R[k][a1] = R[k][a2] | R[k][a3]; break;
	case XOR: 	R[k][a1] = R[k][a2] ^ R[k][a3]; break;
	case EQ: 	R[k][a1] = R[k][a2] == R[k][a3]; break;
	case NE: 	R[k][a1] = R[k][a2] != R[k][a3]; break;
	case LT: 	R[k][a1] = R[k][a2] <  R[k][a3]; break;
	case LE: 	R[k][a1] = R[k][a2] <= R[k][a3]; break;
	case GT: 	R[k][a1] = R[k][a2] >  R[k][a3]; break;
	case GE: 	R[k][a1] = R[k][a2] >= R[k][a3]; break;
	case SHL: 	R[k][a1] = R[k][a2] << R[k][a3]; break;
	case SHR: 	R[k][a1] = R[k][a2] >> R[k][a3]; break;
	
	case LDX: 	R[k][a1] = M[ R[k][a2] + R[k][a3]]; break;
	case STX: 	M[ R[k][a2] + R[k][a3]] = R[k][a1]; break;
	case RET: 	Pc[k] = R[k][a1]; break;
	case TRAP: 	trap(k, R[k][a1], a3); break;
	case PUSH: 	R[k][a1]++; M[R[k][a1]] = R[k][a2]; break;
	case POP: 	R[k][a2] = M[R[k][a1]]; R[k][a1]--; break;
 
	default: error("undefine op");
	}
	R[k][0] = 0;
}
/*
void chk_int(int k){
	if( int_flag[k] && int_mask ){
		int_flag[k] = 0;	// clear int
		R[k][31] = Pc;		// save current pc
		Pc[k] = M[10];		// int vector
	}
}
*/
//void display(int k);

// simulate embed sys
void sim(void){
	runflag[0] = 1;
	runflag[1] = 0;                     // start with core 1
	while(runflag[0] || runflag[1]){
		if(runflag[0]) run(0);
		if(runflag[1]) run(1);
//		timer(0);	// run timer
//		chk_int();
		if( ninst[0] > XLIMIT) exit(0);
		if( ninst[1] > XLIMIT) exit(0);
	}
//	display(0);
//	display(1);
}

int main(int argc, char *argv[]){
	if( argc < 2 ) {
		printf("usage : s23 objfile\n");
		exit(0);
	}
	loadprogram(argv[1]);
//	dumpcode();
	DP = HEAP;
	clock = 0;
	ninst[0] = 0;
	ninst[1] = 0;
	Pc[0] = 100;
//	Pc[1] = 100;
	R[0][0] = 0;
	R[1][0] = 0;
//	int_mask = 0;  	// disable int
//	int_flag = 0;  	// no int
//	timer(1);		// init timer
	sim();
//	shell();
	return 0;
}

